# -*- coding: utf-8 -*-

import numpy as np

def kinase_classification(pfam_domain_path):
    tyr_kinase=[]
    st_kinase=[]
    f1=open(pfam_domain_path)
    seq=f1.readline()
    seq=seq.strip().split("\t")
    tyr_kinase.extend(seq[1:])
    seq=f1.readline()
    seq=seq.strip().split("\t")
    st_kinase.extend(seq[1:])
    return tyr_kinase,st_kinase
    
def load_seeds(pfam_domain_path, sim_mean_std_path, sim_all_folder_path, 
               test_path, graph_nodes):
    '''
    Loading seed nodes.
    '''
    # zscores_global.
    f1 = open(sim_mean_std_path)
    seq=f1.readline()
    seq=f1.readline()
    zscores_global={}
    while(seq!=""):
        seq=seq.strip().split("\t")
        zscores_global[seq[0]]=np.array(seq[1].split("|"),dtype=float)
        seq=f1.readline()

    # seed nodes.
    f1 = open(test_path) 
    seq=f1.readline()
    ssim={}
    seeds_pos={}
    seeds_neg={}
    while(seq!=""):
        seq= seq.strip().split("\t")
        seq[1]=float(seq[1])
        if seq[1]>0.0:
            seeds_pos[seq[0]]=seq[1]
        if seq[1]<-0.0:
            seeds_neg[seq[0]]=-seq[1]
        seq=f1.readline()
    
    '''
    Classification and semantic similarity.
    '''
    # Seed node classification.
    tyr_kinase,st_kinase=kinase_classification(pfam_domain_path)
    tyr_pos=list((set(tyr_kinase).intersection(seeds_pos.keys())).intersection(set(graph_nodes)))
    tyr_neg=list((set(tyr_kinase).intersection(seeds_neg.keys())).intersection(set(graph_nodes)))
    st_pos= list((set(st_kinase).intersection(seeds_pos.keys())).intersection(set(graph_nodes)))
    st_neg=list((set(st_kinase).intersection(seeds_neg.keys())).intersection(set(graph_nodes)))
    all_the_rest_pos=list((set(seeds_pos.keys()).difference(set(tyr_pos+st_pos))).intersection(set(graph_nodes)))
    all_the_rest_neg=list((set(seeds_neg.keys()).difference(set(tyr_neg+st_neg))).intersection(set(graph_nodes)))
    seeds=[tyr_pos,st_pos,all_the_rest_pos,tyr_neg,st_neg,all_the_rest_neg]

    # Semantic similarity for each seed node.
    ssim={}
    for i in list(set(tyr_pos+tyr_neg+st_pos+st_neg+all_the_rest_pos+all_the_rest_neg)):
        ssim[i]={}
        fsim=open(sim_all_folder_path+'/'+i+"_all.txt")
        # The first readline trim away the header.
        seq_sim=fsim.readline()
        seq_sim=fsim.readline()
        while(seq_sim!=""):
            seq_sim=seq_sim.strip().split("\t")
            ssim[i][seq_sim[1]]=float(seq_sim[2])
            seq_sim=fsim.readline()
            
    '''
    Output.
    '''
    return seeds_pos,seeds_neg,seeds,zscores_global,ssim


# -*- coding: utf-8 -*-

from .utils_CLI import dataprep
from .utils_CLI import load_test_example
from .phuego import phuego
from .phuego_mock import phuego_mock

from .load_seeds import load_seeds
from .network_rwr import rwr_values
from .network_rwr import pvalue_split
from .ego import ego_filtering
from .ego2module import merge_egos
from .generate_net import generate_nets



# -*- coding: utf-8 -*-

import numpy as np
import igraph as ig
from .utils import fisher_test

def rwr_values(network, graph_nodes, ini_pos, ini_neg, seeds, seeds_pos, 
               seeds_neg, network_path, network_random_path, damping, res_folder):
    
    number_of_nodes=network.vcount()
    empirical_rwr=np.zeros((6,number_of_nodes),dtype=float)

    empirical_values={}
    pvalues={}
    for i in network.vs:
        pvalues[i["name"]]=np.zeros(6,dtype=int)
        empirical_values[i["name"]]=np.zeros(6,dtype=float)

    pos=False
    if ini_pos[0]!="False":
        pos=True
        to_delete=[]
        for j in network.vs.select(name_in=ini_pos):
            to_delete.append(j.index)

    neg=False
    if ini_neg[0]!="False":
        neg=True
        to_delete=[]
        for j in network.vs.select(name_in=ini_neg):
            to_delete.append(j.index)
    flag_pos=0
    flag_neg=0
    for i in enumerate(seeds):
        if len(i[1])>0:
            if pos==True and flag_pos==0 and i[0]<=2:
                network.delete_vertices(to_delete)
                flag_pos=1
            if neg==True and flag_neg==0 and i[0]>2:
                network.delete_vertices(to_delete)
                flag_neg=1

            number_of_nodes=network.vcount()
            reset_vertex=np.zeros(number_of_nodes)

            for j in network.vs.select(name_in=i[1]):
                if i[0]<=2:
                    reset_vertex[j.index]=seeds_pos[j["name"]]
                else:
                    reset_vertex[j.index]=seeds_neg[j["name"]]
            pagerank=np.array(network.personalized_pagerank(reset=reset_vertex,directed=False, damping=damping, weights='weight',implementation='prpack'))
            if flag_pos==1 and i[0]<=2:
                for j in to_delete:
                    pagerank=np.concatenate((pagerank[:j], [0], pagerank[j:]))

            if flag_neg==1 and i[0]>2:
                for j in to_delete:
                    pagerank=np.concatenate((pagerank[:j], [0], pagerank[j:]))

            empirical_rwr[i[0]]=pagerank

        if i[0]==2 and pos==True:
            network = ig.Graph.Read_Ncol(network_path, weights=True, directed=False)
            pos=False
            flag_pos=0
            to_delete=[]


    for i in enumerate(empirical_rwr.T):
        empirical_values[graph_nodes[i[0]]]=i[1]

    for ii in range(1000):
        network_random = ig.Graph.Read_Ncol(network_random_path+str(ii)+".txt", weights=True, directed=False)
        nodes=network_random.vs["name"]
        number_of_nodes=network_random.vcount()
        random_rwr=np.zeros((6,number_of_nodes),dtype=float)
        pos=False
        if ini_pos[0]!="False":
            to_delete=[]
            for j in network_random.vs.select(name_in=ini_pos):

                to_delete.append(j.index)
            pos=True
        neg=False
        if ini_neg[0]!="False":
            to_delete=[]
            for j in network_random.vs.select(name_in=ini_neg):
                to_delete.append(j.index)
            neg=True
        flag_pos=0
        flag_neg=0

        for jj in enumerate(seeds):
            if len(jj[1])>0:
                if pos==True and flag_pos==0 and jj[0]<=2:
                    network_random.delete_vertices(to_delete)
                    flag_pos=1
                if neg==True and flag_neg==0 and jj[0]>2:
                    network_random.delete_vertices(to_delete)
                    flag_neg=1

                number_of_nodes=network_random.vcount()
                reset_vertex=np.zeros(number_of_nodes)

                for j in network_random.vs.select(name_in=jj[1]):
                    if jj[0]<=2:
                        reset_vertex[j.index]=seeds_pos[j["name"]]
                    else:
                        reset_vertex[j.index]=seeds_neg[j["name"]]
                prandom=np.array(network_random.personalized_pagerank(reset=reset_vertex,directed=False, damping=damping, weights='weight',implementation='prpack'))

                if flag_pos==1 and jj[0]<=2:
                    for j in to_delete:
                        prandom=np.concatenate((prandom[:j], [0], prandom[j:]))

                if flag_neg==1 and jj[0]>2:
                    for j in to_delete:
                        prandom=np.concatenate((prandom[:j], [0], prandom[j:]))

                if jj[0]==2 and pos==True:
                    network_random = ig.Graph.Read_Ncol(network_random_path+str(ii)+".txt", weights=True, directed=False)
                    flag_pos=0
                    pos=False
                    to_delete=[]

                random_rwr[jj[0]]=prandom
            else:
                if jj[0]==2:
                    network_random = ig.Graph.Read_Ncol(network_random_path+str(ii)+".txt", weights=True, directed=False)
                    flag_pos=0
                    pos=False
                    to_delete=[]
        random_rwr=random_rwr.T
        for i in enumerate(nodes):
            pvalues[i[1]]=np.add(pvalues[i[1]],np.greater(empirical_values[i[1]],random_rwr[i[0]]))

    '''
    Output.
     '''
    #writing the results for global propagation
    f1=open(res_folder+"rwr_scores.txt","w")
    f1.write('uniprotid\tincreased_tyr\tincreased_kinases\tincreased_substrates\tdecreased_tyr\tdecreased_kinases\tdecreased_substrates\n')
    for i in empirical_values:
        f1.write(i+"\t"+"\t".join(map(str,empirical_values[i]))+"\n")
    f1.close()
    
    f1=open(res_folder+"pvalues.txt","w")
    f1.write('uniprotid\tincreased_tyr\tincreased_kinases\tincreased_substrates\tdecreased_tyr\tdecreased_kinases\tdecreased_substrates\n')
    for i in pvalues:
        f1.write(i+"\t"+"\t".join(map(str,pvalues[i]))+"\n")
    f1.close()
    
    f1=open(res_folder+"start_seeds.txt","w")
    f1.write('uniprotid\tLFC_value\n')
    for i in seeds[0]+seeds[1]+seeds[2]:
        f1.write(i+"\t"+str(seeds_pos[i])+"\n")

    for i in seeds[3]+seeds[4]+seeds[5]:
        f1.write(i+"\t"+str(-seeds_neg[i])+"\n")
    f1.close()
 
    # # Returned variables.
    # return empirical_values,pvalues


# Obtain the significant rwr nodes according to pvalues.txt, and split into increased/decreased nodes.
def pvalue_split(res_folder, seeds, graph_nodes, rwr_threshold,
                 fisher_threshold, fisher_geneset, uniprot_to_gene, geneset_path):
    '''
    Separate the pvalues for upregulated/downregulated nodes.
    '''
    pvalues_pos=[]
    pvalues_neg=[]
 
    f1=open(res_folder+"pvalues.txt")
    seq=f1.readline()
    seq=f1.readline()
    #the pvalues need to be a parameter for the users  
    while(seq!=""):
        seq=seq.strip().split("\t")
        seq[1:]=np.array(seq[1:],float)
        if max(seq[1:4]) > 1000 - rwr_threshold*1000:
            pvalues_pos.append(seq[0])
        if max(seq[4:]) > 1000 - rwr_threshold*1000:
            pvalues_neg.append(seq[0])
        seq=f1.readline()

    pvalues_pos=list(set(pvalues_pos+seeds[0]+seeds[1]+seeds[2]))
    pvalues_neg=list(set(pvalues_neg+seeds[3]+seeds[4]+seeds[5]))
    #perfrom fishertest on the seeds and rwr nodes
    
    fname = "increased_rwr_fisher"
    fisher_test(protein_list=pvalues_pos,
                starting_proteins=list(set(seeds[0]+seeds[1]+seeds[2])),
                fname=fname,
                threshold=fisher_threshold,
                component=fisher_geneset,
                path_def=res_folder,
                uniprot_to_gene=uniprot_to_gene,
                geneset_path=geneset_path,
                )
    fname = "decreased_rwr_fisher"
    fisher_test(protein_list=pvalues_neg,
            starting_proteins=list(set(seeds[3]+seeds[4]+seeds[5])),
            fname=fname,
            threshold=fisher_threshold,
            component=fisher_geneset,
            path_def=res_folder,
            uniprot_to_gene=uniprot_to_gene,
            geneset_path=geneset_path,
            )
    fname = "increased_seed_fisher"
    fisher_test(protein_list=list(set(seeds[0]+seeds[1]+seeds[2])),
            starting_proteins=list(set(seeds[0]+seeds[1]+seeds[2])),
            fname=fname,
            threshold=fisher_threshold,
            component=fisher_geneset,
            path_def=res_folder,
            uniprot_to_gene=uniprot_to_gene,
            geneset_path=geneset_path,
        )
    fname = "decreased_seed_fisher"
    fisher_test(protein_list=list(set(seeds[3]+seeds[4]+seeds[5])),
            starting_proteins=list(set(seeds[3]+seeds[4]+seeds[5])),
            fname=fname,
            threshold=fisher_threshold,
            component=fisher_geneset,
            path_def=res_folder,
            uniprot_to_gene=uniprot_to_gene,
            geneset_path=geneset_path,
        )

    
    '''
    Separate the rwr_scores for upregulated/downregulated nodes.
    '''
    f1=open(res_folder+"rwr_scores.txt")
    rwr_pos={}
    rwr_neg={}
    seq=f1.readline()
    seq=f1.readline()
    while(seq!=""):
        seq=seq.strip().split("\t")
        rwr_pos[seq[0]]=np.mean(np.array(seq[1:4],float))
        rwr_neg[seq[0]]=np.mean(np.array(seq[4:],float))
        seq=f1.readline()

    '''
    Force node assignment to be unique (either up- or down-regulated).
    '''
    common=list(set(pvalues_pos).intersection(pvalues_neg))
    remove_from_pos=[]
    remove_from_neg=[]
    for i in common:
        if rwr_pos[i]<rwr_neg[i]:
            remove_from_pos.append(i)
        if rwr_neg[i]<rwr_pos[i]:
            remove_from_neg.append(i)

    pvalues_pos=list(set(pvalues_pos).difference(remove_from_pos))
    pvalues_neg=list(set(pvalues_neg).difference(remove_from_neg))

    pvalues_pos=list(set(pvalues_pos).intersection(graph_nodes))
    pvalues_neg=list(set(pvalues_neg).intersection(graph_nodes))
    
    return(pvalues_pos, pvalues_neg)



# -*- coding: utf-8 -*-

from .utils import denoise_square
from .utils import calc_kde
from .utils import fisher_test
import numpy as np
import igraph
from scipy.spatial.distance import jensenshannon

def ego_filtering(network, pval, seeds, sim, zscores_global, kde_cutoff, 
                  direction, uniprot_to_gene, res_folder,
                  geneset_path, fisher_geneset, fisher_threshold, ):
    # Below: only add kde_cutoff to ego_friends input.
    subnet = network.induced_subgraph(list(network.vs.select(name_in=pval)),implementation='create_from_scratch')
    subnet.vs.select(_degree=0).delete()
    nodes=subnet.vs["name"]
    n_nodes=subnet.vcount()
    position_nodes={k: v for v, k in enumerate(nodes)}
    seed_nodes=list(set(seeds).intersection(set(nodes)))
    nodes_kde=ego_friends(subnet=subnet,
                          position_nodes=position_nodes,
                          seed_nodes=seed_nodes,
                          sim=sim,
                          zscores=zscores_global,
                          kde_cutoff=kde_cutoff,
                          )
    
    # Fisher test and write results.
    supernodes,all_nodes=write_results(nodes_kde=nodes_kde,
                                       seed_nodes=seed_nodes, 
                                       kde_cutoff=kde_cutoff, 
                                       direction=direction, 
                                       uniprot_to_gene=uniprot_to_gene, 
                                       res_folder=res_folder,
                                       geneset_path=geneset_path, 
                                       fisher_geneset=fisher_geneset, 
                                       fisher_threshold=fisher_threshold,
                                       )
    return supernodes,all_nodes

def ego_friends(subnet,position_nodes,seed_nodes,sim,zscores,kde_cutoff):

    nodes_kde={}
    for i in kde_cutoff:
        nodes_kde[i]={}

    for i in enumerate(sorted(seed_nodes)):
        ego=subnet.induced_subgraph(subnet.neighborhood(vertices=position_nodes[i[1]], order=2, mode='all', mindist=0),implementation='create_from_scratch')
        nodes=ego.vs["name"]
        n_nodes=len(nodes)
        if n_nodes>5:
            position_ego={k: v for v, k in enumerate(nodes)}
            functional_nodes=[position_ego[i[1]]]

            for j in nodes:
                num=sim[i[1]][j]-float(zscores[j][0])
                den=float(zscores[j][1])
                if den>0 and (num/den)>1.645:
                    if i[1]!=j:
                        functional_nodes.append(position_ego[j])
            ego=ego.induced_subgraph(functional_nodes,implementation='copy_and_delete')
            ego=ego.induced_subgraph(ego.neighborhood(vertices=ego.vs.find(i[1]), order=2, mode='all', mindist=0),implementation='copy_and_delete')
            ego.vs.select(_degree=0).delete()
            nodes=ego.vs["name"]
            n_nodes=len(nodes)
            position_ego={k: v for v, k in enumerate(nodes)}
            if len(ego.connected_components(mode='strong'))>1 and i[1] in position_ego:
                ego=ego.induced_subgraph(ego.subcomponent(position_ego[i[1]], mode='all'),implementation='copy_and_delete')
                ego.vs.select(_degree=0).delete()
                nodes=ego.vs["name"]
                n_nodes=len(nodes)
                position_ego={k: v for v, k in enumerate(nodes)}

            if i[1] in position_ego and n_nodes>5:
                position_ego={k: v for v, k in enumerate(nodes)}
                first_shell=ego.neighbors(position_ego[i[1]], mode='all')

                distances=dict.fromkeys(first_shell,1)
                nodes_id=ego.vs.indices
                second_shell=list(set(nodes_id).difference(set(first_shell+[position_ego[i[1]]])))
                for j in second_shell:
                    distances[j]=2
                distances[position_ego[i[1]]]=0
                for j in ego.es:
                    node_A=j.tuple[0]
                    node_B=j.tuple[1]
                    dist=[distances[node_A],distances[node_B]]
                    if (dist[0]==1 and dist[1]==1) or (dist[0]==2 and dist[1]==2):
                        ego_sim= (sim[i[1]][nodes[node_A]]+sim[i[1]][nodes[node_B]])/2.0
                        ego.es[j.index]["weight"]=ego_sim
                    elif (dist[0]==1 and dist[1]==2):
                        ego_sim=sim[i[1]][nodes[node_B]]
                        ego.es[j.index]["weight"]=ego_sim
                    elif (dist[0]==2 and dist[1]==1):
                        ego_sim=sim[i[1]][nodes[node_A]]
                        ego.es[j.index]["weight"]=ego_sim
                ego=denoise_square(ego)

                reset_vertex=np.zeros(n_nodes)
                reset_vertex[position_ego[i[1]]]=1.0
                ego_rwr=np.array(ego.personalized_pagerank(reset=reset_vertex,directed=False, damping=0.85, weights='weight',implementation='prpack'))
                ssim=[]
                dist=[]

                for j in ego.vs:
                    reset_vertex=np.zeros(n_nodes)
                    reset_vertex[j.index]=1.0
                    ego_node=np.array(ego.personalized_pagerank(reset=reset_vertex,directed=False, damping=0.85,implementation='prpack',weights='weight'))
                    if i[1]==j["name"]:
                        dist.append(1.0)
                    else:
                        dist.append(1-jensenshannon(ego_node,ego_rwr))
                    ssim.append(sim[i[1]][j["name"]])

                ssim=np.array(ssim)
                dist=np.array(dist)
                ssim=1000*np.log2(1+ssim)
                dist=1000*np.log2(1+dist)

                cdf_dist,grid_dist=calc_kde(dist)
                cdf_ssim,grid_ssim=calc_kde(ssim)

                for j in kde_cutoff:
                    position=np.searchsorted(cdf_dist,j)
                    interpolation=np.argmin([abs(j-cdf_dist[position]),abs(j-cdf_dist[position-1])])
                    grid_threshold_dist=grid_dist[position-interpolation]

                    position=np.searchsorted(cdf_ssim,j)
                    interpolation=np.argmin([abs(j-cdf_ssim[position]),abs(j-cdf_ssim[position-1])])
                    grid_threshold_ssim=grid_ssim[position-interpolation]

                    nodes_sim=[i[1]]
                    nodes_dist=[i[1]]
                    for k in enumerate(nodes):
                        if ssim[k[0]]>grid_threshold_ssim:
                            nodes_sim.append(k[1])
                        if dist[k[0]]>grid_threshold_dist:
                            nodes_dist.append(k[1])
                    nodes_kde[j][i[1]]=list(set(nodes_sim).intersection(nodes_dist))

    return nodes_kde

def write_results(nodes_kde, seed_nodes, kde_cutoff, direction, uniprot_to_gene, 
                  res_folder, geneset_path, fisher_geneset, fisher_threshold, ):
    all_nodes={}
    supernodes={}
    for j in kde_cutoff:
        all_nodes[j]=set()
        supernodes[j]={}

        f2=open(res_folder+direction+"_sig"+"_cluster_"+str(j)+".txt", "w")
        fisher_proteins=set()
        for k in nodes_kde[j]:
            f2.write(k+"\t"+"\t".join(nodes_kde[j][k])+"\n")
            if len(nodes_kde[j][k])>=2:
                supernodes[j][k]=nodes_kde[j][k]
                fisher_proteins.update(nodes_kde[j][k])
                
        if fisher_proteins:
            # Create filename for fisher output.
            fname = direction+"_sig_fisher_"+str(j)
            fisher_test(protein_list=fisher_proteins,
                        starting_proteins=seed_nodes,
                        fname=fname,
                        threshold=fisher_threshold,
                        component=fisher_geneset,
                        path_def=res_folder,
                        uniprot_to_gene=uniprot_to_gene,
                        geneset_path=geneset_path,
                        )
            all_nodes[j]=list(fisher_proteins)
    return supernodes,all_nodes



# -*- coding: utf-8 -*-

import igraph as ig
import sys
from .utils import load_gene_names
from .utils import add_trailing_slash
from .utils import convert_result
from .load_seeds import load_seeds
from .network_rwr import rwr_values
from .network_rwr import pvalue_split
from .ego import ego_filtering
from .ego2module import merge_egos
from .generate_net import generate_nets



# one liner function of phuego package.
def phuego(support_data_folder, res_folder, test_path, 
           fisher_geneset, fisher_threshold, fisher_background,
           ini_pos, ini_neg, damping, kde_cutoff,rwr_threshold,
           use_existing_rwr = False, convert2folder = False, 
           include_isolated_egos_in_KDE_net=False,
           net_format="ncol",):
       
    """
    Formatting user input.
    """
    # Force the folder paths to end with /.
    support_data_folder = add_trailing_slash(support_data_folder)
    res_folder = add_trailing_slash(res_folder)
        
    # Force damping to be [0.5, 0.95]
    if type(damping) is not float:
           sys.exit("damping should be a float number within range [0.5, 0.95]")
    if damping < 0.5 or damping > 0.95:
           sys.exit("damping should be a float number within range [0.5, 0.95]")
    # Force rwr_threshold to be [0.01, 0.1]
    if type(rwr_threshold) is not float:
           sys.exit("rwr_threshold should be a float number within range [0.01, 0.1]")
    if rwr_threshold > 0.1 or rwr_threshold < 0.01:
           sys.exit("rwr_threshold should be a float number within range [0.01, 0.1]")
    
    # Force kde_cutoff input to be list and within [0.5, 0.95]
    if type(kde_cutoff) is not list:
           kde_cutoff = [ kde_cutoff ]
    for kde in kde_cutoff:
           if kde < 0.5 or kde > 0.95:
                  sys.exit("kde_cutoff should be within range [0.5, 0.95]")

    # Force the fisher_geneset input to be list.
    if type(fisher_geneset) is not list:
           fisher_geneset = [ fisher_geneset ]
    for geneset in fisher_geneset:
           if geneset not in ["C","F","D","P","R","K","RT","B"]:
                  sys.exit("fisher genesets should be chosen from C,F,D,P,R,K,RT,B")

    """
    Create support data paths.
    """       
    # Networks
    network_path = support_data_folder + "networks/gic.txt"
    network_raw_path = support_data_folder + "networks/gic_raw.txt"
    network_random_path = support_data_folder + "networks/gic_random/"
    # Protein annotation, ID mapping
    pfam_domain_path = support_data_folder + "pfam_domains.txt"
    uniprot_to_gene_path = support_data_folder + "uniprot_to_gene.tab"
    # Semantic similarity
    sim_mean_std_path = support_data_folder + "gic_mean_std.txt"
    sim_all_folder_path = support_data_folder + "gic_sim/"
    # Genesets.
    if(fisher_background == "proteome"):
       geneset_path = support_data_folder + "proteome/"
    elif(fisher_background == "intact"):
       geneset_path = support_data_folder + "intact/"
    else:
       sys.exit("Please provide a valid fisher background, either proteome or intact.")

    """
    load network.
    """
    network = ig.Graph.Read_Ncol(network_path, weights=True, directed=False)
    graph_nodes=network.vs["name"]
    
    network_raw = ig.Graph.Read_Ncol(network_raw_path, weights=True, directed=False)
    graph_nodes_raw = network_raw.vs["name"]
    
    """
    Load uniprot ID to gene name mapping.
    """
    uniprot_to_gene=load_gene_names(uniprot_to_gene_path)
    
    """
    load seeds.
    """
    seeds_pos,seeds_neg,seeds,zscores_global,ssim = load_seeds(
                     pfam_domain_path=pfam_domain_path,
                     sim_mean_std_path=sim_mean_std_path,
                     sim_all_folder_path=sim_all_folder_path,
                     test_path=test_path,
                     graph_nodes=graph_nodes,
                     )
    
    """
    Random walk with restart
    """
    # On genuine and randomized netowrk.
    # TIME CONSUMING: skip and use existing result by 'use_existing_rwr = True'
    if(use_existing_rwr):
           pass
    else:
           rwr_values(network=network, 
                      graph_nodes=graph_nodes, 
                      ini_pos=ini_pos,
                      ini_neg=ini_neg,
                      seeds=seeds,
                      seeds_pos=seeds_pos,
                      seeds_neg=seeds_neg,
                      network_path=network_path,
                      network_random_path=network_random_path,
                      damping=damping,
                      res_folder=res_folder,
                     )
    
    """
    Split the pvalues.
    """
    # pvalues is imported from file in res_folder.
    pvalues_pos, pvalues_neg = pvalue_split(res_folder=res_folder,
                                          seeds=seeds,
                                          graph_nodes=graph_nodes,
                                          fisher_threshold=fisher_threshold,
                                          fisher_geneset=fisher_geneset,
                                          uniprot_to_gene=uniprot_to_gene,
                                          geneset_path=geneset_path,
                                          rwr_threshold=rwr_threshold,
                                          )
    
    """
    EGO.
    """
    if pvalues_pos:
           nodes_kde_pos, all_nodes_pos = ego_filtering(network=network,
                                   pval=pvalues_pos,
                                   seeds=seeds_pos,
                                   sim=ssim,
                                   zscores_global=zscores_global,
                                   kde_cutoff=kde_cutoff,
                                   direction="increased",
                                   uniprot_to_gene=uniprot_to_gene,
                                   res_folder=res_folder,
                                   geneset_path=geneset_path,
                                   fisher_geneset=fisher_geneset,
                                   fisher_threshold=fisher_threshold,
                                   )
    if pvalues_neg:
           nodes_kde_neg, all_nodes_neg = ego_filtering(network=network,
                                   pval=pvalues_neg,
                                   seeds=seeds_neg,
                                   sim=ssim,
                                   zscores_global=zscores_global,
                                   kde_cutoff=kde_cutoff,
                                   direction="decreased",
                                   uniprot_to_gene=uniprot_to_gene,
                                   res_folder=res_folder,
                                   geneset_path=geneset_path,
                                   fisher_geneset=fisher_geneset,
                                   fisher_threshold=fisher_threshold,
                                   )
    
    """
    Module.
    """  
    try:
       if all_nodes_pos:
              merge_egos(network=network_raw,
                            kde_cutoff=kde_cutoff,
                            res_folder=res_folder,
                            uniprot_to_gene=uniprot_to_gene,
                            direction="increased",
                            supernodes=nodes_kde_pos,
                            all_nodes=all_nodes_pos,
                            geneset_path=geneset_path,
                            fisher_geneset=fisher_geneset,
                            fisher_threshold=fisher_threshold,
                            )
    except NameError:
       print(f"No node is significantly propagated during rwr of increased seed nodes. To confirm, length of pvalues_pos is {len(pvalues_pos)}")

    try:       
       if all_nodes_neg:
              merge_egos(network=network_raw,
                            kde_cutoff=kde_cutoff,
                            res_folder=res_folder,
                            uniprot_to_gene=uniprot_to_gene,
                            direction="decreased",
                            supernodes=nodes_kde_neg,
                            all_nodes=all_nodes_neg,
                            geneset_path=geneset_path,
                            fisher_geneset=fisher_geneset,
                            fisher_threshold=fisher_threshold,
                            )
    except NameError:
       print(f"No node is significantly propagated during rwr of decreased seed nodes. To confirm, length of pvalues_neg is {len(pvalues_neg)}")
           
    """
    Output CytoScape compatible network files.
    """
    generate_nets(res_folder=res_folder,
                  network=network_raw,
                  uniprot_to_gene=uniprot_to_gene,
                  kde_cutoff=kde_cutoff,
                  rwr_threshold=rwr_threshold,
                  include_isolated_egos_in_KDE_net=include_isolated_egos_in_KDE_net,
                  net_format=net_format,
                  )
    
    """
    Convert results if required.
    """
    if(convert2folder):
           convert_result(res_folder=res_folder,
                          kde_cutoff=kde_cutoff,
                          net_format=net_format,
                          )



import igraph as ig
import numpy as np
import pandas as pd
import os

def graph_to_df(G, seed, nodes_modules):
    edges = []
    # Network file as annotated edge list.
    for e in G.es:
        ProteinA = G.vs[e.source]["name"]
        ProteinA_GeneName = G.vs[e.source]["Gene_name"]
        ProteinB = G.vs[e.target]["name"]
        ProteinB_GeneName = G.vs[e.target]["Gene_name"]
        edge_dict = {
            "ProteinA": ProteinA,
            "ProteinB": ProteinB,
            "ProteinA_GeneName": ProteinA_GeneName,
            "ProteinB_GeneName": ProteinB_GeneName,
            "weight": e["weight"],
            "A_is_seed": ProteinA in seed,
            "B_is_seed": ProteinB in seed,
            "all_modules": []
        }
        in_any_module = False
        # Check if the edge is in the modules.
        for module_name, node_list in sorted(nodes_modules.items()):
            in_module = (ProteinA in node_list) and (ProteinB in node_list)
            edge_dict[f"is_{module_name}"] = in_module
            if in_module:
                in_any_module = True
                edge_dict["all_modules"].append(module_name.split("_")[1])
        edge_dict["inter_module"] = not in_any_module
        edges.append(edge_dict)
    df_edges = pd.DataFrame(edges)
    
    # Network node attribute file.
    nodes = []
    for v in G.vs:
        Protein = v["name"]
        Protein_GeneName = v["Gene_name"]
        node_dict = {
            "Protein": Protein,
            "Protein_GeneName": Protein_GeneName,
            "Protein_is_seed": Protein in seed,
        }
        for module_name, node_list in sorted(nodes_modules.items()):
            in_module = (Protein in node_list)
            node_dict[f"is_{module_name}"] = in_module
        nodes.append(node_dict)
    df_nodes = pd.DataFrame(nodes)
    return df_edges, df_nodes

def generate_nets(res_folder, network, uniprot_to_gene, kde_cutoff, rwr_threshold,
                  include_isolated_egos_in_KDE_net,net_format,):
    if os.path.isdir(res_folder):
        gene_label=[]
        for i in network.vs["name"]:
            gene_label.append(uniprot_to_gene.get(i,"Not_available"))
        network.vs["Gene_name"]=gene_label

        f1=open(res_folder+"start_seeds.txt")
        seq=f1.readline()
        seq=f1.readline()
        seeds_increase=[]
        seeds_decrease=[]
        while(seq!=""):
            seq= seq.strip().split("\t")
            seq[1]=float(seq[1])
            if seq[1]>0.0:
                seeds_increase.append(seq[0])
            if seq[1]<-0.0:
                seeds_decrease.append(seq[0])
            seq=f1.readline()
        f1.close()

        f1=open(res_folder+"pvalues.txt")
        pvalues_pos=[]
        pvalues_neg=[]
        seq=f1.readline()
        seq=f1.readline()
        while(seq!=""):
            seq=seq.strip().split("\t")
            seq[1:]=np.array(seq[1:],float)
            if max(seq[1:4]) > 1000 - rwr_threshold*1000:
                pvalues_pos.append(seq[0])
            if max(seq[4:]) > 1000 - rwr_threshold*1000:
                pvalues_neg.append(seq[0])
            seq=f1.readline()
        f1.close()
  
        """
        RWR NETWORKS -- down/upregulated.
        """
        pvalues_pos=pvalues_pos+seeds_increase
        pvalues_neg=pvalues_neg+seeds_decrease
        rwr_net_increased=network.induced_subgraph(pvalues_pos)
        rwr_net_increased["title"] ="RWR_increased_net"
        number_of_nodes=rwr_net_increased.vcount()
        rwr_net_increased.vs["Is_seed"]=list(np.full(number_of_nodes, False))
        for i in rwr_net_increased.vs:
            if i["name"] in seeds_increase:
                rwr_net_increased.vs[i.index]["Is_seed"]=True
        # Output the increased network.
        fname = res_folder+"rwr_increased."+net_format
        ig.write(rwr_net_increased,fname,format=net_format)


        rwr_net_decreased=network.induced_subgraph(pvalues_neg)
        rwr_net_decreased["title"] ="RWR_decreased_net"

        number_of_nodes=rwr_net_decreased.vcount()
        rwr_net_decreased.vs["Is_seed"]=list(np.full(number_of_nodes, False))
        for i in rwr_net_decreased.vs:
            if i["name"] in seeds_decrease:
                rwr_net_decreased.vs[i.index]["Is_seed"]=True
        # Output the decreased network.
        fname = res_folder+"rwr_decreased."+net_format
        ig.write(rwr_net_decreased,fname,format=net_format)


        for i in kde_cutoff:
            # The kde is a float. Convert to string for using in path.
            i = str(i)
            # Depending on the input, the sig_cluster file might not exist for increased seed nodes.
            if os.path.isfile(res_folder+"increased_sig_cluster_"+i+".txt"):
                """
                UPREGULATED -- SIGNATURE NETWORK.
                """
                f1=open(res_folder+"increased_sig_cluster_"+i+".txt")
                seq=f1.readline()
                nodes=set()
                while(seq!=""):
                    seq=seq.strip().split("\t")
                    if include_isolated_egos_in_KDE_net==False:
                        if len(seq)>2:
                            nodes.update(seq)
                    else:
                        nodes.update(seq)
                    seq=f1.readline()
                KDE_increased=rwr_net_increased.induced_subgraph(nodes)
                KDE_increased["title"] ="KDE_increased_net"
                number_of_nodes=KDE_increased.vcount()
                ##write the net
                fname = res_folder+"KDE_increased_"+i+"."+net_format
                ig.write(KDE_increased,fname,format=net_format)
   
                """
                UPREGULATED -- MODULE NETWORKS
                """ 
                files = os.listdir(res_folder)
                module_files = [file_name for file_name in files 
                                if ("increased_module_" in file_name) and 
                                    (("_cluster_"+i+".txt") in file_name)]
                # Get the module name, such as "module_0".
                modules = []
                for file_name in module_files:
                    module = file_name.split("_")[2]
                    module = "module_" + module
                    modules.append(module)

                all_nodes=set()
                nodes_module=dict()
                for module, module_file in zip(modules, module_files): 
                    KDE_increased.vs[module]=list(np.full(number_of_nodes, False))
                    f1 = open(res_folder+module_file)
                    seq=f1.readline()
                    nodes_module[module]=set()
                    while(seq!=""):
                        seq=seq.strip().split("\t")
                        nodes_module[module].update(seq)
                        seq=f1.readline()
                    # Label the KDE_increased vertices attribute with module name.
                    # This vertice attribute will be inherited to module_net when subgraph is induced.
                    KDE_increased.vs[module]=np.in1d(KDE_increased.vs["name"], list(nodes_module[module]))
                    all_nodes.update(nodes_module[module])
                #write the net
                module_net=KDE_increased.induced_subgraph(all_nodes)
                module_net["title"] ="Module_increased_net"
                fname = res_folder+"module_net_increased_"+i+"."+net_format
                ig.write(module_net,fname,format=net_format)
                
                # Create the dataframe for annotated csv output of module network.
                seed = seeds_increase + seeds_decrease
                df_module_net_edges, df_module_net_nodes = graph_to_df(module_net, seed, nodes_module)
                df_module_net_edges.to_csv(res_folder+"module_net_increased_edgelist_"+i+".csv")
                df_module_net_nodes.to_csv(res_folder+"module_net_increased_nodes_attribute_"+i+".csv")
            else:
                pass

            # Depending on the input, the sig_cluster file might not exist for decreased seed nodes.
            if os.path.isfile(res_folder+"decreased_sig_cluster_"+i+".txt"):
                """
                DOWNREGULATED -- SIGNATURE NETWORK
                """
                f1=open(res_folder+"decreased_sig_cluster_"+i+".txt")
                seq=f1.readline()
                nodes=set()
                while(seq!=""):
                    seq=seq.strip().split("\t")
                    if include_isolated_egos_in_KDE_net==False:
                        if len(seq)>2:
                            nodes.update(seq)
                    else:
                        nodes.update(seq)
                    seq=f1.readline()
                KDE_decreased=rwr_net_decreased.induced_subgraph(nodes)
                KDE_decreased["title"] ="KDE_decreased_net"
                number_of_nodes=KDE_decreased.vcount()
                ##write the net
                fname = res_folder+"KDE_decreased_"+i+"."+net_format
                ig.write(KDE_decreased,fname,format=net_format)

                """
                DOWNREGULATED -- MODULE NETWORKS
                """ 
                # Get the module file names (for downregulated and a specific kde_cutoff)
                files = os.listdir(res_folder)
                module_files = [file_name for file_name in files 
                                if ("decreased_module_" in file_name) and 
                                    (("_cluster_"+i+".txt") in file_name)]
                # Get the module name, such as "module_0".
                modules = []
                for file_name in module_files:
                    module = file_name.split("_")[2]
                    module = "module_" + module
                    modules.append(module)
                
                all_nodes=set()
                nodes_module=dict()
                for module, module_file in zip(modules, module_files):
                    KDE_decreased.vs[module]=list(np.full(number_of_nodes, False))
                    f1=open(res_folder+module_file)
                    seq=f1.readline()
                    nodes_module[module]=set()
                    while(seq!=""):
                        seq=seq.strip().split("\t")
                        nodes_module[module].update(seq)
                        seq=f1.readline()
                    all_nodes.update(nodes_module[module])
                    KDE_decreased.vs[module]=np.in1d(KDE_decreased.vs["name"], list(nodes_module[module]))

                #write the net
                module_net=KDE_decreased.induced_subgraph(all_nodes)
                module_net["title"] ="Module_decreased_net"
                fname = res_folder+"module_net_decreased_"+i+"."+net_format
                ig.write(module_net,fname,format=net_format)
                
                # Create the dataframe for annotated csv output of module network.
                seed = seeds_increase + seeds_decrease
                df_module_net_edges, df_module_net_nodes = graph_to_df(module_net, seed, nodes_module)
                df_module_net_edges.to_csv(res_folder+"module_net_decreased_edgelist_"+i+".csv")
                df_module_net_nodes.to_csv(res_folder+"module_net_decreased_nodes_attribute_"+i+".csv")
            else:
                pass



# -*- coding: utf-8 -*-

import requests
import hashlib
import tarfile
import os
import sys
import pandas as pd

def dataprep(support_data_folder, need_fisher=True, need_gic_sim=True, need_networks=True, 
             remove_zip_file = False):
    
    # Input retrieved from Zenodo. Should be static unless files are changed by developer.
    project_url = "https://zenodo.org/api/records/8094690"
    fisher_md5 = "eb68bd6fd16ae340afccef390787ca90"
    gic_sim_md5 = "fa608665e4e6ce2241f5b5dcf105c552"
    networks_md5 = "6c249ac5c049815eb7a3fadcbb5c149c"
    
    # Send a GET request to the download URL
    response = requests.get(project_url)
    if response.status_code == 200:
        data = response.json()
        fisher_url = data["files"][0]["links"]["self"]
        gic_sim_url = data["files"][1]["links"]["self"]
        networks_url = data["files"][2]["links"]["self"]
    else:
        sys.exit("Cannot access Zenodo. Please troubleshoot.")
        
    if(need_fisher):
        download_dataset(fisher_url, support_data_folder, fisher_md5, remove_zip_file)
    if(need_gic_sim):
        download_dataset(gic_sim_url, support_data_folder, gic_sim_md5, remove_zip_file)
    if(need_networks):
        download_dataset(networks_url, support_data_folder, networks_md5, remove_zip_file)


def download_dataset(download_url, support_data_folder, file_md5, remove_zip_file):
    # Send a GET request to the download URL
    response = requests.get(download_url)

    if response.status_code == 200:
        # Extract the filename from the response headers
        filename = response.headers.get("Content-Disposition").split("=")[1]
        
        # Save the dataset to a file
        with open(support_data_folder+filename, "wb") as file:
            file.write(response.content)
        
        # Check if md5_checksum matches. If so decompress file.
        md5_checksum = calculate_md5(support_data_folder+filename)
        if(file_md5 == md5_checksum):
            decompress_tar_gz(support_data_folder+filename, support_data_folder)
            print(f"MD5 checksum match, dataset '{filename}' downloaded successfully!")
        else:
            sys.exit("MD5 checksum doesn't match, download incomplete.")

        # Remove the tar.gz file after decompression, if required by the user.    
        if(remove_zip_file):
            os.remove(support_data_folder+filename)
            print(f"{filename} has been removed after decompression.")
    else:
        sys.exit("Cannot access Zenodo. Please troubleshoot.")
        

def calculate_md5(file_path):
    md5_hash = hashlib.md5()
    with open(file_path, "rb") as file:
        # Read the file in chunks to conserve memory
        for chunk in iter(lambda: file.read(4096), b""):
            md5_hash.update(chunk)
    return md5_hash.hexdigest()


def decompress_tar_gz(file_path, output_dir):
    with tarfile.open(file_path, "r:gz") as tar:
        tar.extractall(path=output_dir)       

    
def load_test_example():
    test_path = os.path.join(os.path.dirname(__file__), 'data/EGF_vs_Untreated_@min_15_63_240.txt')
    df = pd.read_csv(test_path, sep='\t', header=None)
    return test_path, df



# -*- coding: utf-8 -*-

from .utils_CLI import load_test_example
from .utils_CLI import dataprep
from .phuego import phuego
from .phuego_mock import phuego_mock
import click

__version__ = "1.0.0"

'''
This is the CLI tool for phuego.
'''
@click.command()
# Folders.
@click.option("--support_data_folder", "-sf", type=str, required=False, 
              help="Folder that stored support data. Directories separated by /")
@click.option("--res_folder", "-rf", type=str, required=False, 
              help="Folder that stored output result. Directories separated by /")
@click.option("--test_path", "-tpath", default="", type=str, required=False, 
              help="The path of user input test file. If defined, -mock and -test should not be used.")
@click.option("--ini_path", "-ipath", default="", type=str, required=False, 
              help="The path of a csv file, that specify two sets of nodes to be removed from reference network during propagation.")
@click.option("--convert2folder/--dont_convert2folder", default=True, required=False, 
              help="Should phuego organize result files into a folder structure?")


# Support dataset download.
@click.option("--need_fisher", is_flag=True, required=False,
              help="Should phuego download the fisher geneset from zenodo?")
@click.option("--need_gic_sim", is_flag=True, required=False, 
              help="Should phuego download the semantic similarity from zenodo?")
@click.option("--need_networks", is_flag=True, required=False, 
              help="Should phuego download the networks from zenodo?")
@click.option("--remove_zip_file", is_flag=True, required=False, 
              help="Should phuego remove the zip files after downloading?")

# Mock/test.
@click.option("--run_mock", is_flag=True, required=False, 
              help="Should phuego perform a mock test run using test data? If defined, -test and -tpath would be ignored.")
@click.option("--run_test", is_flag=True, required=False, 
              help="Should phuego perform a full test run using test data? If defined, -tpath would be ignored.")


# network propagation parameters.
@click.option("--use_existing_rwr", "-ru", is_flag=True, required=False, 
              help="Should phuego reuse existing network propagation results?")
@click.option("--damping", "-d", default=0.85, type=float, required=False, 
              help="Damping factor for random walk with restart algorithm. Float number within range [0.5, 0.95]")
@click.option("--rwr_threshold", "-rt", default=0.05, type=float, required=False, 
              help="Threshold of significance for propagated nodes. Float number within range [0.01, 0.1]")

# ego propagation and fisher test.
@click.option("--kde_cutoff", "-k", default=[0.85], type=float, multiple=True,
              required=False, 
              help="KDE cutoff value for removing less similar nodes in ego network. A float number within range [0.5, 0.95]. Multiple numbers can be provided at the same time by reusing the argument flag.")
@click.option("--fisher_geneset", "-fg", default=["K"], type=str, multiple=True,
              required=False, 
              help="Abbreviation of genesets to be tested when annotating modules of propagation, can be one of 'C,F,D,P,R,K,RT,B'. Multiple genesets can be provided at the same time by reusing the argument flag. Refer to documents to learn what each abbreviation stands for.")
@click.option("--fisher_threshold", "-ft", default=0.05, type=float, 
              required=False, 
              help="Threshold of significance of geneset enrichment analysis") 
@click.option("--fisher_background", "-fb", default="intact", type=str, 
              required=False, 
              help="Geneset background of geneset enrichment analysis") 

# network output
@click.option("--include_isolated_egos_in_kde_net", "-ie", is_flag=True, required=False, 
              help="Should we include isolated nodes in the output network?") 
@click.option("--net_format", "-nf", default="graphml", type=str, required=False, 
              help="file format of output network. Can be one of 'edgelist', 'pajek', 'ncol', 'lgl', 'graphml', 'dimacs', 'gml', 'dot', 'leda'")

# Versioning.
@click.option('--version', '-v', is_flag=True, 
              help='Print version to stdout')

def main(support_data_folder, res_folder, test_path, convert2folder, use_existing_rwr, run_test, run_mock, 
         need_fisher, need_gic_sim, need_networks, remove_zip_file,
         damping, fisher_geneset, fisher_threshold, fisher_background, kde_cutoff, ini_path, rwr_threshold, 
         include_isolated_egos_in_kde_net, net_format, version) -> None:
       
    # Print the version number.
    if version:
        bold_package_name = click.style("phuego", bold=True)
        click.echo(f"{bold_package_name} ({__version__})")
        return

    # Download support dataset.
    if(need_fisher | need_gic_sim | need_networks):
       dataprep(support_data_folder=support_data_folder, 
              need_fisher=need_fisher, 
              need_gic_sim=need_gic_sim, 
              need_networks=need_networks,
              remove_zip_file=remove_zip_file)
       

    # Click turns the multiple input into tuple. Make it as list now.
    fisher_geneset = list(fisher_geneset)
    kde_cutoff = list(kde_cutoff)
    # ini_pos = list(ini_pos)
    # ini_neg = list(ini_neg)
        
    # Create ini_pos and ini_neg, by default or by reading the ini_path file. 
    if(ini_path == ""):
       ini_pos = ["False"]
       ini_neg = ["False"]
    else:
       f1=open(ini_path)
       ini_pos = f1.readline().strip()
       ini_pos = ini_pos.split(",")
       ini_neg = f1.readline().strip()
       ini_neg = ini_neg.split(",")

    # Run phuego.
    if(run_mock):
           test_path, test_df = load_test_example()
           print("Run phuego_mock with test dataset, whose first few lines are: \n",test_df.head())
           phuego_mock(
              support_data_folder=support_data_folder,
              res_folder=res_folder,
              test_path=test_path,
              fisher_geneset=fisher_geneset,
              fisher_threshold=fisher_threshold,
              fisher_background=fisher_background,
              ini_pos=ini_pos,
              ini_neg=ini_neg,
              damping=damping,
              kde_cutoff=kde_cutoff,
              use_existing_rwr=use_existing_rwr,
              convert2folder=convert2folder,
              include_isolated_egos_in_KDE_net=include_isolated_egos_in_kde_net,
              net_format=net_format,
              )
    elif(run_test):
           test_path, test_df = load_test_example()
           print("Run phuego with test dataset, whose first few lines are: \n",test_df.head())
           phuego(
              support_data_folder=support_data_folder,
              res_folder=res_folder,
              test_path=test_path,
              fisher_geneset=fisher_geneset,
              fisher_threshold=fisher_threshold,
              fisher_background=fisher_background,
              ini_pos=ini_pos,
              ini_neg=ini_neg,
              damping=damping,
              rwr_threshold=rwr_threshold,
              kde_cutoff=kde_cutoff,
              use_existing_rwr=use_existing_rwr,
              convert2folder=convert2folder,
              include_isolated_egos_in_KDE_net=include_isolated_egos_in_kde_net,
              net_format=net_format,
              )
    elif(test_path):
           print("Run phuego with user input dataset")
           phuego(
              support_data_folder=support_data_folder,
              res_folder=res_folder,
              test_path=test_path,
              fisher_geneset=fisher_geneset,
              fisher_threshold=fisher_threshold,
              fisher_background=fisher_background,
              ini_pos=ini_pos,
              ini_neg=ini_neg,
              damping=damping,
              rwr_threshold=rwr_threshold,
              kde_cutoff=kde_cutoff,
              use_existing_rwr=use_existing_rwr,
              convert2folder=convert2folder,
              include_isolated_egos_in_KDE_net=include_isolated_egos_in_kde_net,
              net_format=net_format,
              )
    else:
           print("Nothing is provided for running phuego.")


# -*- coding: utf-8 -*-

from .utils import denoise_square
from .utils import fisher_test
import igraph as ig
import numpy as np
import leidenalg as la
from itertools import combinations
from scipy.spatial.distance import jensenshannon


def merge_egos(network, kde_cutoff, res_folder, uniprot_to_gene,
                  supernodes,all_nodes,direction,geneset_path,fisher_geneset, 
                  fisher_threshold,):
    for kde in kde_cutoff:        
        kde_net=network.induced_subgraph(network.vs.select(name_in=all_nodes[kde]),implementation='copy_and_delete')
        supernodes_net=ig.Graph()
        temp=[]
        for i in combinations(supernodes[kde].keys(),2):
            subnet=kde_net.induced_subgraph(kde_net.vs.select(name_in=supernodes[kde][i[0]]+supernodes[kde][i[1]]),implementation='copy_and_delete')
            subnet.vs.select(_degree=0).delete()
            subnet=denoise_square(subnet)
            nodes=subnet.vs["name"]
            n_nodes=len(nodes)
            index_net={k: v for v, k in enumerate(nodes)}
            components=subnet.connected_components()

            if i[0] in index_net and i[1] in index_net and components.membership[index_net[i[0]]]==components.membership[index_net[i[1]]]:
                reset_vertex=np.zeros(n_nodes)
                reset_vertex[index_net[i[0]]]=1.0
                pagerankA=np.array(subnet.personalized_pagerank(reset=reset_vertex,directed=False, damping=0.85, weights='weight',implementation='prpack'))

                reset_vertex=np.zeros(n_nodes)
                reset_vertex[index_net[i[1]]]=1.0
                pagerankB=np.array(subnet.personalized_pagerank(reset=reset_vertex,directed=False, damping=0.85, weights='weight',implementation='prpack'))
                js_dist=jensenshannon(pagerankA,pagerankB)

                if js_dist>0:
                    temp.append((i[0],i[1],js_dist))
        supernodes_net = ig.Graph.TupleList(temp,weights="weight",directed = False)
        components = supernodes_net.connected_components(mode="strong")
        modules=[]
        mod={}
        for i in enumerate(components):
            if len(i[1])>=4:
                cc_net=supernodes_net.induced_subgraph(i[1],implementation='auto')
                nodes=cc_net.vs["name"]
                cluster = la.find_partition(cc_net, la.ModularityVertexPartition,weights='weight',n_iterations=-1,seed=42)
                for jj in cluster:
                    modules.append(cc_net.vs.select(jj)["name"])

            elif 2<=len(i[1])<4:
                sm_net=supernodes_net.induced_subgraph(i[1],implementation='auto')
                nodes=sm_net.vs["name"]
                modules.append(nodes)
        
        fname = direction+"_supernodes_net_"+str(kde)+".txt"
        supernodes_net.write(f=res_folder+fname,format="ncol")
        write_modules(clustering=modules,
                      nodes=supernodes[kde],
                      kde=kde,
                      direction=direction,
                      uniprot_to_gene=uniprot_to_gene,
                      res_folder=res_folder,
                      geneset_path=geneset_path,
                      fisher_geneset=fisher_geneset,
                      fisher_threshold=fisher_threshold,
                     )

def write_modules(clustering,nodes, kde, direction, uniprot_to_gene, res_folder, 
                  geneset_path, fisher_geneset, fisher_threshold, ):
    for j in enumerate(clustering):
        f1=open(res_folder+direction+"_module_"+str(j[0])+"_cluster_"+str(kde)+".txt", "w")
        fisher_proteins=set()
        for k in j[1]:
            f1.write(k+"\t"+"\t".join(nodes[k])+"\n")
            fisher_proteins.update(nodes[k])
        
        fname = direction+"_module_"+str(j[0])+"_fisher_"+str(kde)
        fisher_test(protein_list=fisher_proteins,
                    starting_proteins=j[1],
                    threshold=fisher_threshold,
                    component=fisher_geneset,
                    path_def=res_folder,
                    uniprot_to_gene=uniprot_to_gene,
                    geneset_path=geneset_path,
                    fname=fname,
                    )


# -*- coding: utf-8 -*-

import os
import sys
import shutil
import numpy as np
import math
from sklearn.neighbors import KernelDensity
from scipy.stats import fisher_exact


# Check if folder path end with forward slash, if not, add it.
def add_trailing_slash(folder_path):
    if not folder_path.endswith('/'):
        folder_path = folder_path + '/'
    return folder_path


def load_gene_names(uniprot_to_gene_path):
	uniprot_to_gene={}
	f1=open(uniprot_to_gene_path,"r")
	seq=f1.readline()
	while(seq!=""):
		seq=seq.strip().split("\t")
		uniprot_to_gene[seq[0]]=seq[1].split(";")[0].strip()
		seq=f1.readline()
	return uniprot_to_gene


def denoise_square(G):
	weight=G.strength(G.vs, mode='all', loops=False, weights='weight')
	for i in G.es():
		node_A=i.tuple[0]
		node_B=i.tuple[1]
		den=math.sqrt(weight[node_A]*weight[node_B])
		num=i["weight"]
		G.es[i.index]["weight"]=num/den
	return (G)


def calc_kde(vector):
	obs = len(vector)
	sigma = np.std(vector, ddof=1)
	IQR = (np.percentile(vector, q=75) - np.percentile(vector, q=25)) / 1.3489795003921634
	sigma = min(sigma, IQR)
	if sigma > 0:
		bw= sigma * (obs * 3 / 4.0) ** (-1 / 5)
	else:
		IQR = (np.percentile(vector, q=99) - np.percentile(vector, q=1)) / 4.6526957480816815
		if IQR > 0:
			bw = IQR * (obs * 3 / 4.0) ** (-1 / 5)
	if bw:
		kde = KernelDensity(kernel = 'gaussian', bandwidth=bw).fit(vector.reshape(-1,1))
		grid = np.linspace(min(vector)-1,max(vector)+1,len(vector)*100).reshape(-1,1)
		log_dens = kde.score_samples(grid)
		pdf=np.exp(log_dens)
		grid=grid.ravel()
		normalization=sum(pdf.ravel())
		cdf=np.cumsum(pdf)/normalization
	else:
		onevec=np.ones(len(vector))
		return onevec,onevec

	return cdf,grid


def fisher_test(protein_list,threshold,component,path_def,starting_proteins,
                uniprot_to_gene,geneset_path, fname):
    protein=protein_list
    temp={}
    for ii in component:
        descr={}
        temp[ii]=[]
        f1=open(geneset_path+ii+"_descr.txt","r")
        seq=f1.readline()
        while (seq!=""):
            seq=seq.strip().split("\t")
            descr[seq[0]]=seq[1]
            seq=f1.readline()
        f1.close()
        f1=open(geneset_path+ii+".txt","r")
        seq=f1.readline()
        fisher={}
        fisher_count={}
        while(seq!=""):
            seq=seq.strip().split("\t")
            fisher[seq[0]]=seq[1:]
            seq=f1.readline()
        f1.close()
        f1=open(geneset_path+ii+"_count.txt","r")
        seq=f1.readline()
        while(seq!=""):
            seq=seq.strip().split("\t")
            fisher_count[seq[0]]=int(seq[1])
            seq=f1.readline()
        f1.close()
        fisherset={}
        fisherset_count=[]
        protein_annotation={}
        for i in protein:
            if i in fisher:
                for j in fisher[i]:
                    if j in protein_annotation:
                        protein_annotation[j].append(i)
                    else:
                        protein_annotation[j]=[]
                        protein_annotation[j].append(i)

                    if j in fisherset:
                        fisherset[j]=fisherset[j]+1
                    else:
                        fisherset[j]=1

            else:
                fisherset_count.append(i)


        totalfisher=len(fisher)
        numberofproteins=len(protein)-len(list(set(fisherset_count)))
        fisher={}
        fisher_value_ic={}
        fisher_value={}
        fisher_value_no={}
        lenfisherset=len(fisherset)
        for i in fisherset:

            a=fisherset[i]
            b=numberofproteins-a
            c=fisher_count[i]-a
            d=totalfisher-a-b-c
            table=[[a,b],[c,d]]
            
            # Debugging code in case the negative value bug happen again.
            if a<0 or b<0 or c<0 or d<0:                
                print(f"Current geneset database is:\t{ii}")
                print(f"Current fisherset/geneset is:\t{i}")
                print(f"A:\t{a}")
                print(f"B:\t{b}")
                print(f"C:\t{c}")
                print(f"D:\t{d}")
                print(f"totalfisher:\t{totalfisher}")
                print(f"fisher_count:\t{fisher_count}")
                print(f"fisherset_count:\t{fisherset_count}")
                print(f"numberofproteins:\t{numberofproteins}")
                AAA = len(protein)
                print(f"len(protein):\t{AAA}")
                sys.exit("All values in `table` must be nonnegative.")
    
            fisher[i]=fisher_exact(table,alternative ="greater")[1]
            if fisher[i]<(threshold/lenfisherset):
                if fisher[i] in fisher_value:
                    fisher_value[fisher[i]].append(i)
                else:
                    fisher_value[fisher[i]]=[]
                    fisher_value[fisher[i]].append(i)

        
        # Allow input name.
        f2=open(path_def+fname+"_"+ii+"fisher.txt","w")
        
        # Add a header to the fisher output.
        f2.write("DB id"+"\t"+"adjusted_p"+"\t"+"nodes_in_the_network"+"\t"+"N_of_seed_nodes"+"\t"+"Description"+"\t"+"gene_names"+"\n")
        
        for i in sorted(fisher_value):
            for j in fisher_value[i]:
                temp_gene=[]
                for k in list(set(protein_list).intersection(set(protein_annotation[j]))):
                    temp_gene.append(uniprot_to_gene.get(k,k))
                f2.write(j+"\t"+str(i)+"\t"+str(fisherset[j])+"\t"+str(len(list(set(starting_proteins).intersection(set(protein_annotation[j])))))+"\t"+descr[j]+"\t"+"\t".join(temp_gene)+"\n")


def convert_result(res_folder, kde_cutoff, net_format):
    # Create direction folder and move files.
    files = os.listdir(res_folder)
    for direction in ["decreased", "increased"]:
        # Get file names with direction flag.
        # rwr result files without direction flag will remain in res_folder.
        direction_files = [file_name for file_name in files if direction in file_name]
        
        # Create direction master folder.
        direction_rf = res_folder+direction+"/"
        if os.path.exists(direction_rf):
            print(f"Result sub-folder {direction} already exists from previous run, will be deleted now (along with files)")
            shutil.rmtree(direction_rf)
        os.mkdir(direction_rf)
        # Move rwr network file.
        shutil.move(src=res_folder+"rwr_"+direction+"."+net_format, 
                    dst=direction_rf+"rwr_net."+net_format)
        
        # Create seed fisher folder.
        direction_seed_fisher_rf = direction_rf+"seed_fisher/"
        os.mkdir(direction_seed_fisher_rf)
        # Move all fisher files for seeds.
        srcf = [file_name for file_name in direction_files if "_seed_fisher_" in file_name]
        for file in srcf:
            new_file = file.split("_")[-1]
            shutil.move(src=res_folder+file, dst=direction_seed_fisher_rf+new_file)
            
        # Create rwr fisher folder.
        direction_rwr_fisher_rf = direction_rf+"rwr_fisher/"
        os.mkdir(direction_rwr_fisher_rf)
        # Move all fisher files for rwr.
        srcf = [file_name for file_name in direction_files if "_rwr_fisher_" in file_name]
        for file in srcf:
            new_file = file.split("_")[-1]
            shutil.move(src=res_folder+file, dst=direction_rwr_fisher_rf+new_file)            
        
        
        # Create subfolders and move files.
        for kde in kde_cutoff:
            ######## Create the kde-level folder ########
            direction_KDE_rf = direction_rf+"KDE_"+str(kde)+"/"
            os.mkdir(direction_KDE_rf)
            # Move KDE_egos.txt.
            srcf = [file_name for file_name in direction_files if "_sig_cluster_" in file_name and str(kde) in file_name]
            # there should only be one _sig_cluster_ file per kde x direction.
            if(len(srcf)>1):
                sys.exit(f"More than one _sig_cluster_ file is present for '{direction}'. KDE is {kde}")
            for file in srcf:
                new_file = "KDE_egos.txt"
                shutil.move(src=res_folder+file, dst=direction_KDE_rf+new_file)
            
            ######## Create fisher folder ########
            direction_KDE_fisher_rf = direction_KDE_rf + "fisher/"
            os.mkdir(direction_KDE_fisher_rf)
            # Move all fisher files for signature.
            srcf = [file_name for file_name in direction_files if "_sig_fisher_" in file_name and str(kde) in file_name]
            for file in srcf:
                new_file = file.split("_")[-1]
                shutil.move(src=res_folder+file, dst=direction_KDE_fisher_rf+new_file)
                
            ######## Create network folder ########
            direction_KDE_networks_rf = direction_KDE_rf + "networks/"
            os.mkdir(direction_KDE_networks_rf)
            # Move supernodes_net.
            srcf = [file_name for file_name in direction_files if "_supernodes_net_" in file_name and str(kde) in file_name]
            # There should only be one supernodes_net per kde x direction.
            if(len(srcf)>1):
                sys.exit(f"More than one supernodes_net is present for '{direction}'. KDE is {kde}")
            for file in srcf:
                new_file = "supernodes_net.txt"
                shutil.move(src=res_folder+file, dst=direction_KDE_networks_rf+new_file)
            # Move sig_net and module_net
            srcf = [file_name for file_name in direction_files if net_format in file_name and str(kde) in file_name]
            for file in srcf:
                new_file = file.replace("_"+direction+"_"+str(kde), "")
                shutil.move(src=res_folder+file, dst=direction_KDE_networks_rf+new_file)
            # Move module_net csv file.
            srcf = [file_name for file_name in direction_files if "module_net_" in file_name and ".csv" in file_name and str(kde) in file_name]
            for file in srcf:
                # new_file = "module_net_annotated_edgelist.csv"
                new_file = file.replace("_"+direction, "")
                new_file = new_file.replace("_"+str(kde), "")
                shutil.move(src=res_folder+file, dst=direction_KDE_networks_rf+new_file)
            
            ######## Create module folders ########
            direction_KDE_modules_rf = direction_KDE_rf + "modules/"
            os.mkdir(direction_KDE_modules_rf)
            # Collect the module names for the regulation direction.
            direction_module_names = set()
            for file_name in direction_files:
                if("module_" in file_name and "cluster_"+str(kde) in file_name):
                    module = file_name.split("_")[2]
                    module = "module_" + module
                    direction_module_names.add(module)
            
            # Module folder has extra levels for each module. 
            for module in direction_module_names:
                # Module folder.
                direction_KDE_modules_module_rf = direction_KDE_modules_rf+module+"/"
                os.mkdir(direction_KDE_modules_module_rf)
                # Move module_egos file.
                srcf = [file_name for file_name in direction_files if module in file_name and "_cluster_" in file_name and str(kde) in file_name]
                for file in srcf:
                    new_file = "module_egos.txt"
                    shutil.move(res_folder+file, direction_KDE_modules_module_rf+new_file)
                
                # Module fisher folder.
                direction_KDE_modules_module_fisher_rf = direction_KDE_modules_rf+module+"/fisher/"
                os.mkdir(direction_KDE_modules_module_fisher_rf)
                # Move module fisher files.
                srcf = [file_name for file_name in direction_files if module in file_name and "_fisher_" in file_name and str(kde) in file_name]
                for file in srcf:
                    new_file = file.split("_")[-1]
                    shutil.move(res_folder+file, direction_KDE_modules_module_fisher_rf+new_file)




# -*- coding: utf-8 -*-

import igraph as ig
import numpy as np
import pandas as pd
import sys
import os.path
from .utils import load_gene_names
from .utils import add_trailing_slash
from .utils import convert_result
from .utils import fisher_test
from .load_seeds import load_seeds
from .ego import ego_filtering
from .ego2module import merge_egos
from .generate_net import graph_to_df


# one liner function of phuego package.
def phuego_mock(support_data_folder, res_folder, test_path, 
           fisher_geneset, fisher_threshold, fisher_background,
           ini_pos, ini_neg, damping, kde_cutoff,
           use_existing_rwr = False, convert2folder=False,
           include_isolated_egos_in_KDE_net=False,
           net_format="ncol",):
    """
    Formatting user input.
    """               
    # Force the folder paths to end with /.
    support_data_folder = add_trailing_slash(support_data_folder)
    res_folder = add_trailing_slash(res_folder)
        
    # Force damping to be [0.5, 0.95]
    if type(damping) is not float:
           sys.exit("damping should be a float number within range [0.5, 0.95]")
    if damping < 0.5 or damping > 0.95:
           sys.exit("damping should be a float number within range [0.5, 0.95]")
    
    # Force kde_cutoff input to be list and within [0.5, 0.95]
    if type(kde_cutoff) is not list:
           kde_cutoff = [ kde_cutoff ]
    for kde in kde_cutoff:
           if kde < 0.5 or kde > 0.95:
                  sys.exit("kde_cutoff should be within range [0.5, 0.95]")

    # Force the fisher_geneset input to be list.
    if type(fisher_geneset) is not list:
           fisher_geneset = [ fisher_geneset ]
    for geneset in fisher_geneset:
           if geneset not in ["C","F","D","P","R","K","RT","B"]:
                  sys.exit("fisher genesets should be chosen from C,F,D,P,R,K,RT,B")                  
                  
    """
    Create support data paths.
    """       
    # Networks
    network_path = support_data_folder + "networks/gic.txt"
    network_raw_path = support_data_folder + "networks/gic_raw.txt"
    network_random_path = support_data_folder + "networks/gic_random/"
    # Protein annotation, ID mapping
    pfam_domain_path = support_data_folder + "pfam_domains.txt"
    uniprot_to_gene_path = support_data_folder + "uniprot_to_gene.tab"
    # Semantic similarity
    sim_mean_std_path = support_data_folder + "gic_mean_std.txt"
    sim_all_folder_path = support_data_folder + "gic_sim/"
    # Genesets.
    if(fisher_background == "proteome"):
       geneset_path = support_data_folder + "proteome/"
    elif(fisher_background == "intact"):
       geneset_path = support_data_folder + "intact/"
    else:
       sys.exit("Please provide a valid fisher background, either proteome or intact.")

    """
    load network.
    """
    network = ig.Graph.Read_Ncol(network_path, weights=True, directed=False)
    graph_nodes=network.vs["name"]
    
    network_raw = ig.Graph.Read_Ncol(network_raw_path, weights=True, directed=False)
    graph_nodes_raw = network_raw.vs["name"]
    
    """
    Load uniprot ID to gene name mapping.
    """
    uniprot_to_gene=load_gene_names(uniprot_to_gene_path)
    
    """
    load seeds.
    """
    seeds_pos,seeds_neg,seeds,zscores_global,ssim = load_seeds(
                     pfam_domain_path=pfam_domain_path,
                     sim_mean_std_path=sim_mean_std_path,
                     sim_all_folder_path=sim_all_folder_path,
                     test_path=test_path,
                     graph_nodes=graph_nodes,
                     )
    
    """
    Random walk with restart
    """
    # On genuine and randomized netowrk.
    # TIME CONSUMING: skip and use existing result by 'use_existing_rwr = True'
    if(use_existing_rwr):
           pass
    else:
           rwr_values_mock(network=network, 
                      graph_nodes=graph_nodes, 
                      ini_pos=ini_pos,
                      ini_neg=ini_neg,
                      seeds=seeds,
                      seeds_pos=seeds_pos,
                      seeds_neg=seeds_neg,
                      network_path=network_path,
                      network_random_path=network_random_path,
                      damping=damping,
                      res_folder=res_folder,
                     )
    
    """
    Split the pvalues.
    """
    # pvalues is imported from file in res_folder.
    pvalues_pos, pvalues_neg = pvalue_split_mock(res_folder=res_folder,
                                                 seeds=seeds,
                                                 graph_nodes=graph_nodes,
                                                 fisher_threshold=fisher_threshold,
                                                 fisher_geneset=fisher_geneset,
                                                 uniprot_to_gene=uniprot_to_gene,
                                                 geneset_path=geneset_path,
                                                 )
    
    """
    EGO.
    """
    if pvalues_pos:
           nodes_kde_pos, all_nodes_pos = ego_filtering(network=network,
                                   pval=pvalues_pos,
                                   seeds=seeds_pos,
                                   sim=ssim,
                                   zscores_global=zscores_global,
                                   kde_cutoff=kde_cutoff,
                                   direction="increased",
                                   uniprot_to_gene=uniprot_to_gene,
                                   res_folder=res_folder,
                                   geneset_path=geneset_path,
                                   fisher_geneset=fisher_geneset,
                                   fisher_threshold=fisher_threshold,
                                   )
    if pvalues_neg:
           nodes_kde_neg, all_nodes_neg = ego_filtering(network=network,
                                   pval=pvalues_neg,
                                   seeds=seeds_neg,
                                   sim=ssim,
                                   zscores_global=zscores_global,
                                   kde_cutoff=kde_cutoff,
                                   direction="decreased",
                                   uniprot_to_gene=uniprot_to_gene,
                                   res_folder=res_folder,
                                   geneset_path=geneset_path,
                                   fisher_geneset=fisher_geneset,
                                   fisher_threshold=fisher_threshold,
                                   )
    
    """
    Module.
    """
    if all_nodes_pos:
           merge_egos(network=network_raw,
                     kde_cutoff=kde_cutoff,
                     res_folder=res_folder,
                     uniprot_to_gene=uniprot_to_gene,
                     direction="increased",
                     supernodes=nodes_kde_pos,
                     all_nodes=all_nodes_pos,
                     geneset_path=geneset_path,
                     fisher_geneset=fisher_geneset,
                     fisher_threshold=fisher_threshold,
                     )
           
    if all_nodes_neg:
           merge_egos(network=network_raw,
                     kde_cutoff=kde_cutoff,
                     res_folder=res_folder,
                     uniprot_to_gene=uniprot_to_gene,
                     direction="decreased",
                     supernodes=nodes_kde_neg,
                     all_nodes=all_nodes_neg,
                     geneset_path=geneset_path,
                     fisher_geneset=fisher_geneset,
                     fisher_threshold=fisher_threshold,
                     )

    """
    Output CytoScape compatible network files.
    """
    generate_nets_mock(res_folder=res_folder,
                       network=network_raw,
                       uniprot_to_gene=uniprot_to_gene,
                       kde_cutoff=kde_cutoff,
                       include_isolated_egos_in_KDE_net=include_isolated_egos_in_KDE_net,
                       net_format=net_format,
                       )
    """
    Convert results if required.
    """
    if(convert2folder):
           convert_result(res_folder=res_folder,
                          kde_cutoff=kde_cutoff,
                          net_format=net_format,)



def rwr_values_mock(network, graph_nodes, ini_pos, ini_neg, seeds, seeds_pos, 
               seeds_neg, network_path, network_random_path, damping, res_folder):
    
       number_of_nodes=network.vcount()
       empirical_rwr=np.zeros((6,number_of_nodes),dtype=float)

       empirical_values={}
       pvalues={}
       for i in network.vs:
              pvalues[i["name"]]=np.zeros(6,dtype=int)
              empirical_values[i["name"]]=np.zeros(6,dtype=float)

       pos=False
       if ini_pos[0]!="False":
              pos=True
              to_delete=[]
              for j in network.vs.select(name_in=ini_pos):
                     to_delete.append(j.index)

       neg=False
       if ini_neg[0]!="False":
              neg=True
              to_delete=[]
              for j in network.vs.select(name_in=ini_neg):
                     to_delete.append(j.index)
       flag_pos=0
       flag_neg=0
       for i in enumerate(seeds):
              if len(i[1])>0:
                     if pos==True and flag_pos==0 and i[0]<=2:
                            network.delete_vertices(to_delete)
                            flag_pos=1
                     if neg==True and flag_neg==0 and i[0]>2:
                            network.delete_vertices(to_delete)
                            flag_neg=1

                     number_of_nodes=network.vcount()
                     reset_vertex=np.zeros(number_of_nodes)

                     for j in network.vs.select(name_in=i[1]):
                            if i[0]<=2:
                                   reset_vertex[j.index]=seeds_pos[j["name"]]
                            else:
                                   reset_vertex[j.index]=seeds_neg[j["name"]]
                     pagerank=np.array(network.personalized_pagerank(reset=reset_vertex,directed=False, damping=damping, weights='weight',implementation='prpack'))
                     if flag_pos==1 and i[0]<=2:
                            for j in to_delete:
                                   pagerank=np.concatenate((pagerank[:j], [0], pagerank[j:]))

                     if flag_neg==1 and i[0]>2:
                            for j in to_delete:
                                   pagerank=np.concatenate((pagerank[:j], [0], pagerank[j:]))

                     empirical_rwr[i[0]]=pagerank

              if i[0]==2 and pos==True:
                     network = ig.Graph.Read_Ncol(network_path, weights=True, directed=False)
                     pos=False
                     flag_pos=0
                     to_delete=[]


       for i in enumerate(empirical_rwr.T):
              empirical_values[graph_nodes[i[0]]]=i[1]

       for ii in range(10):
              network_random = ig.Graph.Read_Ncol(network_random_path+str(ii)+".txt", weights=True, directed=False)
              nodes=network_random.vs["name"]
              number_of_nodes=network_random.vcount()
              random_rwr=np.zeros((6,number_of_nodes),dtype=float)
              pos=False
              if ini_pos[0]!="False":
                     to_delete=[]
                     for j in network_random.vs.select(name_in=ini_pos):

                            to_delete.append(j.index)
                     pos=True
              neg=False
              if ini_neg[0]!="False":
                     to_delete=[]
                     for j in network_random.vs.select(name_in=ini_neg):
                            to_delete.append(j.index)
                     neg=True
              flag_pos=0
              flag_neg=0

              for jj in enumerate(seeds):
                     if len(jj[1])>0:
                            if pos==True and flag_pos==0 and jj[0]<=2:
                                   network_random.delete_vertices(to_delete)
                                   flag_pos=1
                            if neg==True and flag_neg==0 and jj[0]>2:
                                   network_random.delete_vertices(to_delete)
                                   flag_neg=1

                            number_of_nodes=network_random.vcount()
                            reset_vertex=np.zeros(number_of_nodes)

                            for j in network_random.vs.select(name_in=jj[1]):
                                   if jj[0]<=2:
                                          reset_vertex[j.index]=seeds_pos[j["name"]]
                                   else:
                                          reset_vertex[j.index]=seeds_neg[j["name"]]
                            prandom=np.array(network_random.personalized_pagerank(reset=reset_vertex,directed=False, damping=damping, weights='weight',implementation='prpack'))

                            if flag_pos==1 and jj[0]<=2:
                                   for j in to_delete:
                                          prandom=np.concatenate((prandom[:j], [0], prandom[j:]))

                            if flag_neg==1 and jj[0]>2:
                                   for j in to_delete:
                                          prandom=np.concatenate((prandom[:j], [0], prandom[j:]))

                            if jj[0]==2 and pos==True:
                                   network_random = ig.Graph.Read_Ncol(network_random_path+str(ii)+".txt", weights=True, directed=False)
                                   flag_pos=0
                                   pos=False
                                   to_delete=[]

                            random_rwr[jj[0]]=prandom
                     else:
                            if jj[0]==2:
                                   network_random = ig.Graph.Read_Ncol(network_random_path+str(ii)+".txt", weights=True, directed=False)
                                   flag_pos=0
                                   pos=False
                                   to_delete=[]
              random_rwr=random_rwr.T
              for i in enumerate(nodes):
                     pvalues[i[1]]=np.add(pvalues[i[1]],np.greater(empirical_values[i[1]],random_rwr[i[0]]))

       '''
       Output.
        '''
       #writing the results for global propagation
       f1=open(res_folder+"rwr_scores.txt","w")
       f1.write('uniprotid\tincreased_tyr\tincreased_kinases\tincreased_substrates\tdecreased_tyr\tdecreased_kinases\tdecreased_substrates\n')
       for i in empirical_values:
              f1.write(i+"\t"+"\t".join(map(str,empirical_values[i]))+"\n")
       f1.close()

       f1=open(res_folder+"pvalues.txt","w")
       f1.write('uniprotid\tincreased_tyr\tincreased_kinases\tincreased_substrates\tdecreased_tyr\tdecreased_kinases\tdecreased_substrates\n')
       for i in pvalues:
              f1.write(i+"\t"+"\t".join(map(str,pvalues[i]))+"\n")
       f1.close()

       f1=open(res_folder+"start_seeds.txt","w")
       f1.write('uniprotid\tLFC_value\n')
       for i in seeds[0]+seeds[1]+seeds[2]:
              f1.write(i+"\t"+str(seeds_pos[i])+"\n")

       for i in seeds[3]+seeds[4]+seeds[5]:
              f1.write(i+"\t"+str(-seeds_neg[i])+"\n")
       f1.close()
 
       # # Returned variables.
       # return empirical_values,pvalues


def pvalue_split_mock(res_folder, seeds, graph_nodes, 
                      fisher_threshold, fisher_geneset, uniprot_to_gene, geneset_path):
    '''
    Separate the pvalues for upregulated/downregulated nodes.
    '''
    pvalues_pos=[]
    pvalues_neg=[]
 
    f1=open(res_folder+"pvalues.txt")
    seq=f1.readline()
    seq=f1.readline()
    while(seq!=""):
        seq=seq.strip().split("\t")
        seq[1:]=np.array(seq[1:],float)
        if max(seq[1:4])>9:
            pvalues_pos.append(seq[0])
        if max(seq[4:])>9:
            pvalues_neg.append(seq[0])
        seq=f1.readline()

    pvalues_pos=list(set(pvalues_pos+seeds[0]+seeds[1]+seeds[2]))
    pvalues_neg=list(set(pvalues_neg+seeds[3]+seeds[4]+seeds[5]))
    
    #perfrom fishertest on the seeds and rwr nodes
    fname = "increased_rwr_fisher"
    fisher_test(protein_list=pvalues_pos,
                starting_proteins=list(set(seeds[0]+seeds[1]+seeds[2])),
                fname=fname,
                threshold=fisher_threshold,
                component=fisher_geneset,
                path_def=res_folder,
                uniprot_to_gene=uniprot_to_gene,
                geneset_path=geneset_path,
                )
    fname = "decreased_rwr_fisher"
    fisher_test(protein_list=pvalues_neg,
            starting_proteins=list(set(seeds[3]+seeds[4]+seeds[5])),
            fname=fname,
            threshold=fisher_threshold,
            component=fisher_geneset,
            path_def=res_folder,
            uniprot_to_gene=uniprot_to_gene,
            geneset_path=geneset_path,
            )
    fname = "increased_seed_fisher"
    fisher_test(protein_list=list(set(seeds[0]+seeds[1]+seeds[2])),
            starting_proteins=list(set(seeds[0]+seeds[1]+seeds[2])),
            fname=fname,
            threshold=fisher_threshold,
            component=fisher_geneset,
            path_def=res_folder,
            uniprot_to_gene=uniprot_to_gene,
            geneset_path=geneset_path,
        )
    fname = "decreased_seed_fisher"
    fisher_test(protein_list=list(set(seeds[3]+seeds[4]+seeds[5])),
            starting_proteins=list(set(seeds[3]+seeds[4]+seeds[5])),
            fname=fname,
            threshold=fisher_threshold,
            component=fisher_geneset,
            path_def=res_folder,
            uniprot_to_gene=uniprot_to_gene,
            geneset_path=geneset_path,
        )
    
    '''
    Separate the rwr_scores for upregulated/downregulated nodes.
    '''
    f1=open(res_folder+"rwr_scores.txt")
    rwr_pos={}
    rwr_neg={}
    seq=f1.readline()
    seq=f1.readline()
    while(seq!=""):
        seq=seq.strip().split("\t")
        rwr_pos[seq[0]]=np.mean(np.array(seq[1:4],float))
        rwr_neg[seq[0]]=np.mean(np.array(seq[4:],float))
        seq=f1.readline()

    '''
    Force node assignment to be unique (either up- or down-regulated).
    '''
    common=list(set(pvalues_pos).intersection(pvalues_neg))
    remove_from_pos=[]
    remove_from_neg=[]
    for i in common:
        if rwr_pos[i]<rwr_neg[i]:
            remove_from_pos.append(i)
        if rwr_neg[i]<rwr_pos[i]:
            remove_from_neg.append(i)

    pvalues_pos=list(set(pvalues_pos).difference(remove_from_pos))
    pvalues_neg=list(set(pvalues_neg).difference(remove_from_neg))

    pvalues_pos=list(set(pvalues_pos).intersection(graph_nodes))
    pvalues_neg=list(set(pvalues_neg).intersection(graph_nodes))
    
    return(pvalues_pos, pvalues_neg)


def generate_nets_mock(res_folder, network, uniprot_to_gene, kde_cutoff,
                  include_isolated_egos_in_KDE_net,net_format,):
    if os.path.isdir(res_folder):
        gene_label=[]
        for i in network.vs["name"]:
            gene_label.append(uniprot_to_gene.get(i,"Not_available"))
        network.vs["Gene_name"]=gene_label

        f1=open(res_folder+"start_seeds.txt")
        seq=f1.readline()
        seq=f1.readline()
        seeds_increase=[]
        seeds_decrease=[]
        while(seq!=""):
            seq= seq.strip().split("\t")
            seq[1]=float(seq[1])
            if seq[1]>0.0:
                seeds_increase.append(seq[0])
            if seq[1]<-0.0:
                seeds_decrease.append(seq[0])
            seq=f1.readline()
        f1.close()

        f1=open(res_folder+"pvalues.txt")
        pvalues_pos=[]
        pvalues_neg=[]
        seq=f1.readline()
        seq=f1.readline()
        while(seq!=""):
            seq=seq.strip().split("\t")
            seq[1:]=np.array(seq[1:],float)
            if max(seq[1:4])>9:
                pvalues_pos.append(seq[0])
            if max(seq[4:])>9:
                pvalues_neg.append(seq[0])
            seq=f1.readline()
        f1.close()
  
        """
        RWR NETWORKS -- down/upregulated.
        """
        pvalues_pos=pvalues_pos+seeds_increase
        pvalues_neg=pvalues_neg+seeds_decrease
        rwr_net_increased=network.induced_subgraph(pvalues_pos)
        rwr_net_increased["title"] ="RWR_increased_net"
        number_of_nodes=rwr_net_increased.vcount()
        rwr_net_increased.vs["Is_seed"]=list(np.full(number_of_nodes, False))
        for i in rwr_net_increased.vs:
            if i["name"] in seeds_increase:
                rwr_net_increased.vs[i.index]["Is_seed"]=True
        # Output the increased network.
        fname = res_folder+"rwr_increased."+net_format
        ig.write(rwr_net_increased,fname,format=net_format)


        rwr_net_decreased=network.induced_subgraph(pvalues_neg)
        rwr_net_decreased["title"] ="RWR_decreased_net"

        number_of_nodes=rwr_net_decreased.vcount()
        rwr_net_decreased.vs["Is_seed"]=list(np.full(number_of_nodes, False))
        for i in rwr_net_decreased.vs:
            if i["name"] in seeds_decrease:
                rwr_net_decreased.vs[i.index]["Is_seed"]=True
        # Output the decreased network.
        fname = res_folder+"rwr_decreased."+net_format
        ig.write(rwr_net_decreased,fname,format=net_format)


        for i in kde_cutoff:
            """
            UPREGULATED -- SIGNATURE NETWORK.
            """
            # The kde is a float. Convert to string for using in path.
            i = str(i)
            f1=open(res_folder+"increased_sig_cluster_"+i+".txt")
            seq=f1.readline()
            nodes=set()
            while(seq!=""):
                seq=seq.strip().split("\t")
                if include_isolated_egos_in_KDE_net==False:
                    if len(seq)>2:
                        nodes.update(seq)
                else:
                    nodes.update(seq)
                seq=f1.readline()
            KDE_increased=rwr_net_increased.induced_subgraph(nodes)
            KDE_increased["title"] ="KDE_increased_net"
            number_of_nodes=KDE_increased.vcount()
            ##write the net
            fname = res_folder+"KDE_increased_"+i+"."+net_format
            ig.write(KDE_increased,fname,format=net_format)
   
            """
            UPREGULATED -- MODULE NETWORKS
            """ 
            files = os.listdir(res_folder)
            module_files = [file_name for file_name in files 
                             if ("increased_module_" in file_name) and 
                                (("_cluster_"+i+".txt") in file_name)]
            # Get the module name, such as "module_0".
            modules = []
            for file_name in module_files:
                module = file_name.split("_")[2]
                module = "module_" + module
                modules.append(module)

            all_nodes=set()
            nodes_module=dict()
            for module, module_file in zip(modules, module_files): 
                KDE_increased.vs[module]=list(np.full(number_of_nodes, False))
                f1 = open(res_folder+module_file)
                seq=f1.readline()
                nodes_module[module]=set()
                while(seq!=""):
                    seq=seq.strip().split("\t")
                    nodes_module[module].update(seq)
                    seq=f1.readline()
                # Label the KDE_increased vertices attribute with module name.
                # This vertice attribute will be inherited to module_net when subgraph is induced.
                KDE_increased.vs[module]=np.in1d(KDE_increased.vs["name"], list(nodes_module[module]))
                all_nodes.update(nodes_module[module])
            #write the net
            module_net=KDE_increased.induced_subgraph(all_nodes)
            module_net["title"] ="Module_increased_net"
            fname = res_folder+"module_net_increased_"+i+"."+net_format
            ig.write(module_net,fname,format=net_format)
            
            # Create the dataframe for annotated csv output of module network.
            seed = seeds_increase + seeds_decrease
            df_module_net_edges, df_module_net_nodes = graph_to_df(module_net, seed, nodes_module)
            df_module_net_edges.to_csv(res_folder+"module_net_increased_edgelist_"+i+".csv")
            df_module_net_nodes.to_csv(res_folder+"module_net_increased_nodes_attribute_"+i+".csv")

            """
            DOWNREGULATED -- SIGNATURE NETWORK
            """
            # The kde is a float. Convert to string for using in path.
            i = str(i)
            f1=open(res_folder+"decreased_sig_cluster_"+i+".txt")
            seq=f1.readline()
            nodes=set()
            while(seq!=""):
                seq=seq.strip().split("\t")
                if include_isolated_egos_in_KDE_net==False:
                    if len(seq)>2:
                        nodes.update(seq)
                else:
                    nodes.update(seq)
                seq=f1.readline()
            KDE_decreased=rwr_net_decreased.induced_subgraph(nodes)
            KDE_decreased["title"] ="KDE_decreased_net"
            number_of_nodes=KDE_decreased.vcount()
            ##write the net
            fname = res_folder+"KDE_decreased_"+i+"."+net_format
            ig.write(KDE_decreased,fname,format=net_format)

            """
            DOWNREGULATED -- MODULE NETWORKS
            """ 
            # Get the module file names (for downregulated and a specific kde_cutoff)
            files = os.listdir(res_folder)
            module_files = [file_name for file_name in files 
                             if ("decreased_module_" in file_name) and 
                                (("_cluster_"+i+".txt") in file_name)]
            # Get the module name, such as "module_0".
            modules = []
            for file_name in module_files:
                module = file_name.split("_")[2]
                module = "module_" + module
                modules.append(module)
            
            all_nodes=set()
            nodes_module=dict()
            for module, module_file in zip(modules, module_files):
                KDE_decreased.vs[module]=list(np.full(number_of_nodes, False))
                f1=open(res_folder+module_file)
                seq=f1.readline()
                nodes_module[module]=set()
                while(seq!=""):
                    seq=seq.strip().split("\t")
                    nodes_module[module].update(seq)
                    seq=f1.readline()
                all_nodes.update(nodes_module[module])
                KDE_decreased.vs[module]=np.in1d(KDE_decreased.vs["name"], list(nodes_module[module]))

            #write the net
            module_net=KDE_decreased.induced_subgraph(all_nodes)
            module_net["title"] ="Module_decreased_net"
            fname = res_folder+"module_net_decreased_"+i+"."+net_format
            ig.write(module_net,fname,format=net_format)
            
            # Create the dataframe for annotated csv output of module network.
            df_module_net_edges, df_module_net_nodes = graph_to_df(module_net, seed, nodes_module)
            df_module_net_edges.to_csv(res_folder+"module_net_decreased_edgelist_"+i+".csv")
            df_module_net_nodes.to_csv(res_folder+"module_net_decreased_nodes_attribute_"+i+".csv")
